#!/usr/bin/env python
# * coding: utf8 *
"""
An example implementation using Supervisor to catch an error and email both the traceback and the logfile
"""

import logging
import logging.handlers
import socket
from pathlib import Path

from supervisor.message_handlers import EmailHandler
from supervisor.models import Supervisor

if __name__ == '__main__':

    #: Set up a rotating file handler for the report log
    test_path = Path(r'c:\temp\supervisor_log.txt')
    test_logger = logging.getLogger('supervisor')
    test_handler = logging.handlers.RotatingFileHandler(test_path, backupCount=2)
    test_handler.doRollover()  #: Rotate the log on each run
    test_handler.setLevel(logging.DEBUG)
    test_logger.addHandler(test_handler)
    test_logger.setLevel(logging.DEBUG)

    #: Add somethign to the log
    test_logger.info('test run')

    #: Instantiate a Supervisor object
    sim_sup = Supervisor('supervisor', logger=test_logger, log_path=test_path)

    #: Specify the email server and addresses
    email_settings = {
        'smtpServer': 'send.state.ut.us',
        'smtpPort': 25,
        'from_address': 'noreply@utah.gov',
        'to_addresses': 'jdadams@utah.gov',
        'prefix': f'Example on {socket.gethostname()}: '
    }

    #: Instantiate a new EmailHandler and register it with our Supervisor
    sim_sup.add_message_handler(EmailHandler(email_settings))

    #: Trigger Supervisor's error handler
    raise ValueError('random error here')



"""
message_handlers.py: Holds all the different message handlers
"""

import gzip
import io
import warnings
from abc import ABC, abstractmethod
from email.mime.application import MIMEApplication
from email.mime.multipart import MIMEMultipart
from email.mime.text import MIMEText
from pathlib import Path
from smtplib import SMTP

import pkg_resources


class MessageHandler(ABC):  # pylint: disable=too-few-public-methods
    """Base class for all message handlers.

    Methods
    -------
    send_message(message_details)
        Parse a MessageDetails object and send a message using handler-specific logic
    """

    @abstractmethod
    def send_message(self, message_details):
        """Send a notification using the target-specific logic (email, slack, etc)

        Parameters
        ----------
        message_details : MessageDetails
            The data to be sent in the notification
        """


class EmailHandler(MessageHandler):  # pylint: disable=too-few-public-methods
    """Send a notification via email

    Attributes
    ----------
    email_settings : dict
        From and To addresses, SMTP Server and Port

    Methods
    -------
    send_message(message_details)
        Build a message, create an SMTP object, and send the message
    _build_message(message_details)
        Create email to be sent as a MIMEMultipart object
    _build_gzip_attachment(input_path)
        gzip input_path into a MIMEApplication object
    """

    def __init__(self, email_settings):
        self.email_settings = email_settings

    def send_message(self, message_details):
        """Build a message, create an SMTP object, and send the message

        Parameters
        ----------
        mesage_details : MessageDetails
            Passed through to _build_message. Must have .message, .subject, .project_name; may
            have .log_file and .attachments
        """

        #: Configure outgoing settings
        try:
            from_address = self.email_settings['from_address']
            to_addresses = self.email_settings['to_addresses']
            smtp_server = self.email_settings['smtpServer']
            smtp_port = self.email_settings['smtpPort']

        except KeyError:
            warnings.warn('Required email settings do not exist. No emails sent.')
            return

        for setting in [from_address, to_addresses, smtp_server, smtp_port]:
            if not setting:
                warnings.warn('Required email settings exist but aren\'t populated. No emails sent.')
                return

        message = self._build_message(message_details)

        #: Send message
        with SMTP(smtp_server, smtp_port) as smtp:
            smtp.sendmail(from_address, to_addresses, message.as_string())

    def _build_message(self, message_details):
        """Create email to be sent as a MIMEMultipart object

        Parameters
        ----------
        mesage_details : MessageDetails
            Must have .message, .subject, .project_name; may have .log_file and .attachments

        Returns
        -------
        message : MIMEMultipart
            A formatted message that can be passed to smtp.sendmail as message.as_string()
        """

        #: Use body as message if it's already a MIMEMultipart, otherwise create a new MIMEMultipart as the message
        message = MIMEMultipart()
        message.attach(MIMEText(message_details.message, 'html'))

        #: Get the client's version (assuming client has been installed via pip install and setup.py)
        distributions = pkg_resources.require(message_details.project_name)
        if distributions:
            version = distributions[0].version
            version = MIMEText(f'<p>{message_details.project_name} version: {version}</p>', 'html')
            message.attach(version)

        #: Split recipient addresses if needed.
        to_addresses = self.email_settings['to_addresses']
        if not isinstance(to_addresses, str):
            to_addresses_joined = ','.join(to_addresses)
        else:
            to_addresses_joined = to_addresses

        #: Add various elements of the message
        if 'prefix' in self.email_settings:
            message['Subject'] = self.email_settings['prefix'] + message_details.subject
        else:
            message['Subject'] = message_details.subject
        message['From'] = self.email_settings['from_address']
        message['To'] = to_addresses_joined

        #: gzip all attachments and add to message
        for original_path in message_details.attachments:
            path = Path(original_path)  #: convert to a Path if it isn't already
            if path.is_file():
                message.attach(self._build_gzip_attachment(path))

        return message

    def _build_gzip_attachment(self, input_path):  #pylint: disable=no-self-use
        """gzip input_path into a MIMEApplication object

        Parameters
        ----------
        input_path : Path
            The on-disk path to the file to gzip. Will be opened in 'rb' mode.

        Returns
        -------
        attachment : MIMEApplication
            The gzip'ed contents of input_path ready to attach to a MIMEMultipart message.
        """
        with (open(input_path, 'rb')) as input_file_object, io.BytesIO() as output_stream:
            gzipper = gzip.GzipFile(mode='wb', fileobj=output_stream)
            gzipper.writelines(input_file_object)
            gzipper.close()
            attachment = MIMEApplication(output_stream.getvalue(), 'x-gzip')
            attachment_filename = input_path.name + '.gz'
            attachment.add_header('Content-Disposition', f'attachment; filename="{attachment_filename}"')

            return attachment


class SlackHandler(MessageHandler):  # pylint: disable=too-few-public-methods
    """
    Send a notification to Slack
    """

    def __init__(self, details):
        pass

    def send_message(self, message_details):
        pass


class ConsoleHandler(MessageHandler):  # pylint: disable=too-few-public-methods
    """Send a notification to the console.

    Methods
    -------
    send_message(message_details)
        Print message_details.message
    """

    def send_message(self, message_details):
        """Print message_details.message"""
        print(message_details.message)



"""
This module holds the classes used by supervisor
"""

import os
import sys
import traceback

from .message_handlers import MessageHandler


class Supervisor:
    """Captures exceptions and sends these and other notifications via email or other methods.

    Attributes
    ----------
    project_name : str
        The name of the client project using Supervisor; used to report version in global exception handler messages
    message_handlers : [MessageHandler]
        Notifications will be sent via all handlers in this list
    logger : Logger
        Logger object to use in global exception handler
    log_path : Path
        Path to the logfile

    Methods
    -------
    add_message_handler(handler)
        Register a new handler for sending messages
    notify(message)
        Send a notification to all registered handlers
    _manage_exceptions
        Closure around a replacement for sys.excepthook
    """

    def __init__(self, project_name, logger=None, log_path=None):

        #: Set up our list of MessageHandlers
        self.project_name = project_name
        self.message_handlers = []
        self.logger = logger
        self.log_path = log_path

        #: Catch any uncaught exception with our custom exception handler
        sys.excepthook = self._manage_exceptions()

    def add_message_handler(self, handler: MessageHandler):
        """Register a new handler for sending messages

        Parameters:
        -----------
        handler : MessageHandler
            A MessageHandler instance for the desired notification method.
        """
        self.message_handlers.append(handler)

    def notify(self, message):
        """Send a notification to all registered handlers by calling .send_message()"""
        for handler in self.message_handlers:
            handler.send_message(message)

    def _manage_exceptions(self):
        """Closure around a replacement for sys.excepthook

        A closure ensures global_exception_handler() has access to self.notify to send notifications while still
        maintaining the signature required by sys.excepthook
        """

        def global_exception_handler(exc_class, exc_object, tb):  # pylint: disable=invalid-name
            """Send a notification via self.notify on any uncaught exceptions; log if client provided a Logger

            Parameters
            ----------
            exc_class : Type
                The type of the exception
            exc_object : Exception
                The exception object
            tb : Traceback
                Original traceback

            Used to handle any uncaught exceptions. Formats an error message, logs it, and sends an email.
            """

            error = os.linesep.join(traceback.format_exception(exc_class, exc_object, tb))

            if self.logger:
                last_traceback = (traceback.extract_tb(tb))[-1]
                line_number = last_traceback[1]
                file_name = last_traceback[0].split('.')[0]
                self.logger.error(f'global error handler line: {line_number} ({file_name})')  # pylint: disable=logging-fstring-interpolation
                self.logger.error(error)

            message_details = MessageDetails()
            message_details.message = error
            message_details.subject = 'ERROR'
            message_details.attachments = [self.log_path]
            message_details.project_name = self.project_name
            self.notify(message_details)

        return global_exception_handler


class MessageDetails:  # pylint: disable=too-few-public-methods
    """Data structure for message details to allow dot-access and null item checks.

    Attributes
    ----------
    message : str
        The text of the message
    attachment : list
        Strings or Paths to any attachments, including log files
    subject : str
        The message subject
    project_name : str
        The name of the project that has added the Supervisor object. Used for adding the version to notifications.

    TODO: Implement true Null-Object pattern.
    """

    def __init__(self):
        self.message = ''
        self.attachments = []  #: Strings or Paths
        self.subject = ''
        self.project_name = ''



#!/usr/bin/env python
# -*- encoding: utf-8 -*-
"""
setup.py
A module that installs supervsior as a module
"""
from glob import glob
from os.path import basename, splitext

from setuptools import find_packages, setup

setup(
    name='agrc-supervisor',
    version='1.1.0',
    license='MIT',
    description='A watchdog module for scheduled scripts that sends notifications, including any uncaught exceptions.',
    author='Jake Adams, AGRC',
    author_email='jdadams@utah.gov',
    url='https://github.com/agrc/supervisor',
    packages=find_packages('src'),
    package_dir={'': 'src'},
    py_modules=[splitext(basename(path))[0] for path in glob('src/*.py')],
    include_package_data=True,
    zip_safe=True,
    classifiers=[
        # complete classifier list: http://pypi.python.org/pypi?%3Aaction=list_classifiers
        'Development Status :: 5 - Production/Stable',
        'Intended Audience :: Developers',
        'Topic :: Utilities',
    ],
    project_urls={
        'Issue Tracker': 'https://github.com/agrc/supervisor/issues',
    },
    keywords=['gis'],
    install_requires=[
        'requests',
    ],
    extras_require={
        'tests': [
            'pylint-quotes==0.2.*',
            'pylint==2.5.*',
            'pytest-cov==2.9.*',
            'pytest-instafail==0.4.*',
            'pytest-isort==1.0.*',
            'pytest-mock',
            'pytest-pylint==0.17.*',
            'pytest-watch==4.2.*',
            'pytest==5.4.*',
            'yapf==0.30.*',
        ]
    },
    setup_requires=[
        'pytest-runner',
    ],
)
